# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-01 18:51
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TaskQueue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scheduled', models.DateTimeField(verbose_name='scheduled')),
                ('module', models.CharField(max_length=256, verbose_name='Module Name')),
                ('function', models.CharField(max_length=256, verbose_name='Function Name')),
                ('arguments', models.BinaryField(blank=True, verbose_name='Arguments')),
                ('is_periodic', models.BooleanField(default=False, verbose_name='Is periodic task')),
                ('timedelta', models.DurationField(blank=True, null=True, verbose_name='timedelta')),
                ('cron_data', models.BinaryField(blank=True, verbose_name='cron data')),
                ('status', models.IntegerField(choices=[(1, 'waiting'), (2, 'running'), (3, 'done'), (4, 'error')], default=1, verbose_name='Status')),
                ('retries', models.IntegerField(default=0, verbose_name='Retries')),
                ('result', models.BinaryField(blank=True, verbose_name='Result')),
                ('ttl', models.DurationField(blank=True, null=True, verbose_name='time to live')),
                ('expire', models.DateTimeField(blank=True, null=True, verbose_name='expire')),
                ('error_message', models.TextField(blank=True, verbose_name='error message')),
            ],
        ),
    ]
